#!/usr/bin/env python
# -*- coding: utf-8 -*-
# wirkungsModell.py
# © Claus Albowski
from wirkungsdiagramme.dialoge import * # eigene

#### Klasse für Faktoren (Groessen) ====================================
class Faktor(object):
    '''Klasse für Faktoren (Groessen)'''

    # Klassenvariable fuer alle Faktornamen und Faktoren
    faktorNamen=[]
    faktoren=[]

    @staticmethod
    def GibAlleFaktorNamen():
        return Faktor.faktorNamen

    @staticmethod
    def GibFaktorZuNamen(name):
        for faktor in Faktor.faktoren:
            if name==faktor.GibNamen():
                return faktor

    @staticmethod
    def LoescheFaktor(name):
        '''loescht den Namen aus der Liste alleFaktoren'''
        try:
            faktor=Faktor.GibFaktorZuNamen(name)
            Wirkung.LoescheWirkungenZuFaktor(name)
            Faktor.faktoren.remove(faktor)
            Faktor.faktorNamen.remove(name)
        except ValueError as e:
            WirkungsModell.GibWirkungsModell().Meldung(str(e))
    
    @staticmethod
    def LoescheAlleFaktoren():
        alleNamen=[]+Faktor.GibAlleFaktorNamen() # echte Kopie!
        for name in alleNamen:
            Faktor.LoescheFaktor(name)

    @staticmethod
    def PruefeNamen(name):
        if name in Faktor.faktorNamen:
            Faktor.GibFaktorZuNamen(name).Meldung("Name existiert bereits!")
            return False
        if name in globals().keys(): return False
        try:
            try:
                auswertung=eval(name)
            except NameError as e:
                # alles ok, keine Zahl
                return True
        except SyntaxError as ee:
            WirkungsModell.GibWirkungsModell().Meldung(str(ee))
            return False
        if auswertung.__class__ in (int,float,complex):
            WirkungsModell.GibWirkungsModell().Meldung("Keine Zahl verwenden!")
            return False
        return True

    
    #### Konstruktor von Faktor
    def __init__(self, name):
        '''fuer die Wirkungsdiagramme werden die Faktoren
        und die Namen der Faktoren gespeichert'''
        while not self.PruefeNamen(name):
            name=input(name+' aendern in: ')
        self.__name=name
        Faktor.faktoren.append(self)
        Faktor.faktorNamen.append(name)

    def GibNamen(self):
        try:
            return self.__name
        except AttributeError as e:
            self.Meldung("Name ist nicht richtig initialisiert:\n"
                         +"SetzeNamen() mit String notwendig!")
    
    def SetzeNeuenNamen(self, neuerName):
        if not(self.PruefeNamen(neuerName)): return False
        alterName=self.GibNamen()
        Faktor.faktorNamen.remove(alterName)
        self.__name=neuerName
        Faktor.faktorNamen.append(neuerName)
        return True

    def Meldung(self, text):
        WirkungsModell.GibWirkungsModell().Meldung(text)

#### Klasse fuer Wirkungen -------------------------------------------
class Wirkung():
    '''Klasse für Wirkungen'''
    
    wirkungen=[]

    @staticmethod
    def HatWirkung(vonName, aufName):
        '''Gibt die Wirkung zurueck, falls sie existiert, sonst False'''
        for wirkung in Wirkung.wirkungen:
            if ((wirkung.GibVon()==vonName)
                and(wirkung.GibAuf()==aufName)): return wirkung
        return False
    
    @staticmethod
    def GibAlleWirkungen():
        return Wirkung.wirkungen
    
    @staticmethod
    def LoescheWirkung(vonName, aufName):
        wirkung=Wirkung.HatWirkung(vonName, aufName)
        if wirkung:
            Wirkung.wirkungen.remove(wirkung)

    @staticmethod
    def LoescheAlleWirkungen():
        wirkungen=[]+Wirkung.GibAlleWirkungen() # echte Kopie
        for wirkung in wirkungen:
            Wirkung.wirkungen.remove(wirkung)
    
    @staticmethod
    def LoescheWirkungenZuFaktor(faktorName):
        '''Faktoren, die entfernt werden sollen, dürfen keine Wirkungen haben.'''
        for wirkung in Wirkung.wirkungen:
            if wirkung.GibVon()==faktorName:
                Wirkung.wirkungen.remove(wirkung)
            if wirkung.GibAuf()==faktorName:
                Wirkung.wirkungen.remove(wirkung)

    #### Konstruktor fuer Wirkungen ------------------------------------    
    def __init__(self, vonName, aufName, art='+'):
        '''Eine Wirkung muss so initialisiert werden, dass sowohl der
        Ausgangs-Faktor als auch der Ziel-Faktor gegeben sind.'''
        self.undefiniert = False
        if Wirkung.HatWirkung(vonName, aufName):
            self.Meldung('Wirkung existiert bereits')
            self.undefiniert = True
        if not(vonName in Faktor.GibAlleFaktorNamen()):
            self.Meldung(vonName+' existiert nicht')
            self.undefiniert = True
        if not(aufName in Faktor.GibAlleFaktorNamen()):
            self.Meldung(aufName+' existiert nicht')
            self.undefiniert = True
        else:
            self.__von=vonName
            self.__auf=aufName
            self.__art=art
            Wirkung.wirkungen.append(self)

    def GibVon(self):
        return self.__von

    def GibAuf(self):
        return self.__auf

    def GibArt(self):
        return self.__art

    def AendereArt(self):
        if self.__art=='+': self.__art='-'
        else: self.__art='+'

    def ZeigeWirkung(self):
        return self.GibVon()+' --> '+self.GibAuf()+' '+self.GibArt()

    def FaktorNameGeaendert(self, alterName,neuerName):
        if self.GibVon()==alterName:
            self.__von=neuerName
        elif self.GibAuf()==alterName:
            self.__auf=neuerName

    def Meldung(self, text):
        WirkungsModell.GibWirkungsModell().Meldung(text)

#### WirkungsModell ----------------------------------------------------------------
class WirkungsModell():
    '''WirkungsModell erstellt das Modell fuer ein Wirkungsdiagramm.'''

    dasModell = None

    @staticmethod
    def GibWirkungsModell(view=None):
        '''Singleton ermoeglichen'''
        if WirkungsModell.dasModell == None:
            WirkungsModell.dasModell = WirkungsModell()
            WirkungsModell.dasModell.SetzeView(view)
        return WirkungsModell.dasModell

    def __init__(self):
        '''Modell fuer ein Wirkungsdiagramm
        Die Faktoren und Wirkungen werden in den Klassen gehalten'''
        self.__modellName=None
        self.__modellDaten=[] # fuer die Ausgabe

    def SetzeModellNamen(self, name):
        assert name.__class__==str, "String eingeben!"
        self.__modellName=name

    def GibModellNamen(self):
        return self.__modellName

    def AendereFaktorNamen(self, alterName, neuerName):
        faktor=Faktor.GibFaktorZuNamen(alterName)
        if faktor.SetzeNeuenNamen(neuerName): ## Wirkungen informieren:
            for wirkung in Wirkung.GibAlleWirkungen():
                wirkung.FaktorNameGeaendert(alterName,neuerName)
            return True
        return False

    def GibAlleFaktorNamen(self):
        return Faktor.GibAlleFaktorNamen()

    def NeuerNameOK(self, name):
        if Faktor.PruefeNamen(name):
            return True
        return False

    def NeuerFaktor(self, name):
        '''Alle Faktoren sind in Faktor gespeichert'''
        if self.NeuerNameOK(name):
            return Faktor(name)
        return None
        
    def NeueWirkung(self, vonName, aufName, art='+'):
        '''Wirkungen werden in der Klasse gehalten'''
        if vonName==aufName: return None
        wirkung=Wirkung(vonName, aufName, art)
        if wirkung.undefiniert: return None
        return wirkung

    def LoescheFaktor(self,name):
        '''entfernt das Faktorobjekt zum Namen
        und entfernt die zugehoerigen wirkungen'''
        faktor=Faktor.GibFaktorZuNamen(name)
        if faktor!=None:
            ## Alle Wirkungen loeschen, die diesen Faktor betreffen.
            for wirkung in Wirkung.GibAlleWirkungen():
                if (name==wirkung.GibVon()) or (name==wirkung.GibAuf()):
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            Faktor.LoescheFaktor(name)
        else:
            self.Meldung("Faktor "+name+" existiert nicht!")
        
    def LoescheWirkung(self, vonName, aufName):
        wirkung=Wirkung.HatWirkung(vonName, aufName)
        if not(wirkung): return False # existiert nicht
        Wirkung.LoescheWirkung(vonName, aufName)

    def LoescheModell(self):
        Wirkung.LoescheAlleWirkungen()
        Faktor.LoescheAlleFaktoren()

#### Abschnitt für die Modelldefinition -----------------------------
    def GibModellDaten(self):
        return self.__modellDaten

    def ZeigeModellDaten(self):
        '''Modelldaten zeilenweise lesbar'''
        daten=self.__modellDaten
        zeilen=['Faktoren:']
        if (len(daten)>0) and (daten[0][0]!='Faktor'):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Faktor'):
            zeilen.append(daten[0][1])
            daten=daten[1:]
        zeilen.append('\nWirkungen:')
        if (len(daten)==0):
            zeilen.append('- - -')
        while (len(daten)>0) and (daten[0][0]=='Wirkung'):
            zeilen.append(daten[0][1]+' -> '+daten[0][2]+' '+daten[0][3])
            daten=daten[1:]
        return zeilen
    
    def SichereModellDaten(self):
        self.__modellDaten=[]
        if self.__modellName==None:
            self.SetzeModellNamen('Neu')
        for name in Faktor.GibAlleFaktorNamen():
            self.__modellDaten.append(('Faktor',name))
        for wirkung in Wirkung.GibAlleWirkungen():
            self.__modellDaten.append(
                ('Wirkung',wirkung.GibVon(),wirkung.GibAuf(),wirkung.GibArt()))
        
    def ErzeugeModellAusDaten(self, modell_Daten):
        for daten in modell_Daten:
            if daten[0]=='Faktor':
                self.NeuerFaktor(daten[1])
            elif daten[0]=='Wirkung':
                self.NeueWirkung(daten[1],daten[2],daten[3])
            else:
                self.Meldung('Fehler bei Modelldefinition: '+daten[0])

    def ModellNameAusDateiNamen(self, dateiname):
        # Pfad abspalten
        if '\\' in dateiname: # windows-Pfadangabe
            liste=dateiname.rsplit('\\')
        else:
            liste=dateiname.rsplit('/')
        liste.reverse()
        name=liste[0]
        if '.' in name:
            name=name.rsplit('.')[0]
        return name

#### Abschnitt für das Laden und Speichern von Modelldaten -----------------
    def Speichere(self, dateiname='', modellDaten=[]):
        self.SichereModellDaten()
        if dateiname=='':
            if self.__modellName==None: return
            dateiname=self.__modellName+'.wirkdia'
        try:
            f=open(dateiname,'w')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten geschrieben')
            return None
        if modellDaten!=[]: f.write(str(modellDaten))
        else: f.write(str(self.__modellDaten))
        f.close()

    def Lies(self, dateiname):
        try:
            f=open(dateiname,'r')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten gelesen')
            return None
        modell_Daten_String=f.read()
        f.close()
        self.__modellName=self.ModellNameAusDateiNamen(dateiname)
        self.__modell_Daten=eval(modell_Daten_String)
        self.ErzeugeModellAusDaten(self.__modell_Daten)
        return self.__modell_Daten

#### Zusaetzliches -----------------------------
    def ZeigeHilfe(self, n=1):
        if n==1:
            hinweise=("TODO"
                      )
        elif n==2:
            hinweise=("TODO.")
        return hinweise
    
    def SetzeView(self, view):
        self.__view=view
        
    def Meldung(self, text):
        if self.__view==None: print(text)
        else:
            MeldungDialog(self.__view,text)
            

#### Tests
##if __name__=='__main__':
##    w=WirkungsModell.GibWirkungsModell()
##    w.NeuerFaktor('Bakterien')
##    w.NeuerFaktor('Zuw_Bakterien')
##    w.NeueWirkung('Zuw_Bakterien','Bakterien')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    wirkung=Wirkung.GibAlleWirkungen()[0]
##    wirkung.AendereArt()
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.AendereFaktorNamen('Bakterien','Bakties')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.LoescheWirkung('Zuw_Bakterien','Bakties')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.NeueWirkung('Zuw_Bakterien','Bakties')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.LoescheFaktor('Bakties')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.NeuerFaktor('Bakterien')
##    w.NeueWirkung('Zuw_Bakterien','Bakterien')
##    w.NeueWirkung('Bakterien','Zuw_Bakterien')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    Wirkung.LoescheAlleWirkungen()    
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    w.NeuerFaktor('1234')
    
